# pylint: disable=duplicate-code

from __future__ import annotations

import copy
import importlib.metadata
import os
import shutil
import sysconfig
import tempfile
import typing
from pathlib import Path
from typing import Any

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from packaging.version import Version

from scikit_build_core.settings.skbuild_model import ScikitBuildSettings

from .._logging import logger, rich_print
from ..build._init import setup_logging
from ..builder.builder import Builder, archs_to_tags, get_archs
from ..builder.get_requires import GetRequires
from ..builder.wheel_tag import WheelTag
from ..cmake import CMake, CMaker
from ..format import pyproject_format
from ..settings.skbuild_read_settings import SettingsReader

__all__ = ["ScikitBuildHook"]


def __dir__() -> list[str]:
    return __all__


class ScikitBuildHook(BuildHookInterface):  # type: ignore[type-arg]
    PLUGIN_NAME = "scikit-build"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__tmp_dir: Path | None = None

    def _read_config(self) -> SettingsReader:
        config_dict = copy.deepcopy(self.config)
        config_dict.pop("dependencies", None)
        config_dict.pop("require-runtime-dependencies", None)
        config_dict.pop("require-runtime-features", None)

        state = typing.cast(
            "typing.Literal['sdist', 'wheel', 'editable']", self.target_name
        )
        return SettingsReader.from_file(
            "pyproject.toml", state=state, extra_settings=config_dict
        )

    def _validate(self, settings_reader: SettingsReader) -> None:
        settings = settings_reader.settings

        settings_reader.validate_may_exit()

        if not settings.experimental:
            msg = "Hatch support is experimental, must enable the experimental flag"
            raise ValueError(msg)

        if not settings.wheel.cmake or settings.sdist.cmake:
            msg = "CMake is required for scikit-build"
            raise ValueError(msg)

        if settings.sdist.include or settings.sdist.exclude:
            msg = "include and exclude are not supported for hatch builds"
            raise ValueError(msg)

        if settings.sdist.cmake:
            msg = "Not currently supported for SDist builds"
            raise ValueError(msg)

        if settings.wheel.packages:
            msg = f"Packages ({settings.wheel.packages!r}) are not supported for hatch builds"
            raise ValueError(msg)

        if (
            settings.wheel.license_files
            and settings.wheel.license_files
            != ScikitBuildSettings().wheel.license_files
        ):
            msg = f"License files ({settings.wheel.license_files!r}) are not supported for hatch builds"
            raise ValueError(msg)

        if settings.wheel.platlib is not None and not settings.wheel.platlib:
            msg = "Purelib builds not supported for hatch builds"
            raise ValueError(msg)

        if settings.generate:
            msg = "Generate is not supported for hatch builds"
            raise ValueError(msg)

        if settings.metadata:
            msg = "Metadata is not supported for hatch builds"
            raise ValueError(msg)

    # Requires Hatchling 1.22.0 to have an effect
    def dependencies(self) -> list[str]:
        settings = self._read_config().settings
        requires = GetRequires(settings)

        if self.target_name == "sdist":
            required = requires.settings.sdist.cmake
        elif self.target_name == "wheel":
            required = requires.settings.wheel.cmake
        else:
            msg = f"Unknown target: {self.target_name!r}, only 'sdist' and 'wheel' are supported"
            raise ValueError(msg)

        # These are only injected if cmake is required
        cmake_requires = [*requires.cmake(), *requires.ninja()] if required else []
        return [*cmake_requires, *requires.dynamic_metadata()]

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        if version == "editable":
            msg = "Editable installs are not yet supported"
            raise ValueError(msg)

        self.__tmp_dir = Path(tempfile.mkdtemp()).resolve()
        try:
            self._initialize(build_data=build_data)
        except Exception:
            self._cleanup()
            raise

    def _initialize(self, *, build_data: dict[str, Any]) -> None:
        settings_reader = self._read_config()
        settings = settings_reader.settings
        state = settings_reader.state

        self._validate(settings_reader)

        if state == "sdist":
            build_data["artifacts"].append("CMakeLists.txt")  # Needs full list, etc.
            return

        setup_logging(settings.logging.level)

        cmake = CMake.default_search(version=settings.cmake.version, env=os.environ)

        rich_print(
            "{green}***",
            "{bold.green}scikit-build-core {__version__}",
            f"using {{blue}}CMake {cmake.version}",
            f"{{red}}({state})",
        )

        self.__tmp_dir = Path(tempfile.mkdtemp()).resolve()
        wheel_dir = self.__tmp_dir / "wheel"

        tags = WheelTag.compute_best(
            archs_to_tags(get_archs(os.environ)),
            settings.wheel.py_api,
            expand_macos=settings.wheel.expand_macos_universal_tags,
            build_tag=settings.wheel.build_tag,
        )
        build_data["tag"] = str(tags)
        build_data["pure_python"] = not settings.wheel.platlib

        build_dir = (
            Path(
                settings.build_dir.format(
                    **pyproject_format(
                        settings=settings,
                        tags=tags,
                        state=state,
                    )
                )
            )
            if settings.build_dir
            else self.__tmp_dir / "build"
        )
        logger.info("Build directory: {}", build_dir.resolve())

        targetlib = "platlib"

        wheel_dirs = {
            targetlib: wheel_dir / targetlib,
            "data": wheel_dir / "data",
            "headers": wheel_dir / "headers",
            "scripts": wheel_dir / "scripts",
            "null": wheel_dir / "null",
            "metadata": wheel_dir / "metadata",
        }

        for d in wheel_dirs.values():
            d.mkdir(parents=True)

        if ".." in settings.wheel.install_dir:
            msg = "wheel.install_dir must not contain '..'"
            raise AssertionError(msg)
        if settings.wheel.install_dir.startswith("/"):
            if not settings.experimental:
                msg = "Experimental features must be enabled to use absolute paths in wheel.install_dir"
                raise AssertionError(msg)
            if settings.wheel.install_dir[1:].split("/")[0] not in wheel_dirs:
                msg = "Must target a valid wheel directory"
                raise AssertionError(msg)
            install_dir = wheel_dir / settings.wheel.install_dir[1:]
        else:
            install_dir = wheel_dirs[targetlib] / settings.wheel.install_dir

        config = CMaker(
            cmake,
            source_dir=settings.cmake.source_dir,
            build_dir=build_dir,
            build_type=settings.cmake.build_type,
        )

        builder = Builder(
            settings=settings,
            config=config,
        )

        rich_print("{green}***", "{bold}Configuring CMake...")
        # Setting the install prefix because some libs hardcode CMAKE_INSTALL_PREFIX
        # Otherwise `cmake --install --prefix` would work by itself
        defines = {"CMAKE_INSTALL_PREFIX": install_dir}
        cache_entries: dict[str, str | Path] = {
            f"SKBUILD_{k.upper()}_DIR": v for k, v in wheel_dirs.items()
        }
        cache_entries["SKBUILD_STATE"] = state
        cache_entries["SKBUILD_HATCHLING"] = importlib.metadata.version("hatchling")
        builder.configure(
            defines=defines,
            cache_entries=cache_entries,
            name=self.build_config.builder.metadata.name,
            version=Version(self.build_config.builder.metadata.version),
        )

        default_gen = (
            "MSVC"
            if sysconfig.get_platform().startswith("win")
            else "Default Generator"
        )
        generator = builder.get_generator() or default_gen
        rich_print(
            "{green}***",
            f"{{bold}}Building project with {{blue}}{generator}{{default}}...",
        )
        build_args: list[str] = []
        builder.build(build_args=build_args)

        rich_print("{green}***", "{bold}Installing project into wheel...")
        builder.install(install_dir)

        files = list(wheel_dirs["headers"].iterdir())
        if files:
            msg = (
                f"Unsupported files found in 'headers' directory: {files}\n"
                "Please report use case to https://github.com/pypa/hatch/issues/1291 if you need it."
            )
            raise ValueError(msg)

        for raw_path in wheel_dirs[targetlib].iterdir():
            path = raw_path.resolve()  # Windows mingw64 and UCRT now requires this
            build_data["force_include"][f"{path}"] = str(
                settings.wheel.install_dir / path.relative_to(wheel_dirs[targetlib])
            )

        try:
            for raw_path in wheel_dirs["data"].iterdir():
                path = raw_path.resolve()  # Windows mingw64 and UCRT now requires this
                build_data["shared_data"][f"{path.resolve()}"] = str(
                    path.relative_to(wheel_dirs["data"])
                )
        except KeyError:
            logger.error("SKBUILD_DATA_DIR not supported by Hatchling < 1.24.0")
            raise

        try:
            for raw_path in wheel_dirs["scripts"].iterdir():
                path = raw_path.resolve()  # Windows mingw64 and UCRT now requires this
                build_data["shared_scripts"][f"{path.resolve()}"] = str(
                    path.relative_to(wheel_dirs["scripts"])
                )
        except KeyError:
            logger.error("SKBUILD_SCRIPTS_DIR not supported by Hatchling < 1.24.0")
            raise

        for raw_path_root in wheel_dirs["metadata"].iterdir():
            path_root = (
                raw_path_root.resolve()
            )  # Windows mingw64 and UCRT now requires this
            if path_root.name != "extra_metadata":
                msg = f"Hatchling metadata must be in an extra_metadata folder, got {path_root}"
                raise ValueError(msg)
            for raw_path in path_root.iterdir():
                path = raw_path.resolve()  # Windows mingw64 and UCRT now requires this
                location = path.relative_to(path_root)
                try:
                    build_data["extra_metadata"][f"{path.resolve()}"] = str(location)
                except KeyError:
                    logger.error("SKBUILD_METADATA_DIR needs a newer Hatchling")
                    raise

    def finalize(
        self, version: str, build_data: dict[str, Any], artifact_path: str
    ) -> None:
        self._cleanup()
        return super().finalize(version, build_data, artifact_path)

    def _cleanup(self) -> None:
        if self.__tmp_dir:
            shutil.rmtree(self.__tmp_dir, ignore_errors=True)
            self.__tmp_dir = None



from __future__ import annotations

import importlib.metadata
import sys
import typing

if typing.TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from importlib.metadata import EntryPoint

        EntryPoints = typing.List[EntryPoint]
    else:
        from importlib.metadata import EntryPoints

__all__ = ["entry_points"]


def entry_points(*, group: str) -> EntryPoints:
    if sys.version_info >= (3, 10):
        return importlib.metadata.entry_points(group=group)

    epg = importlib.metadata.entry_points()
    return epg.get(group, [])  # pylint: disable=no-member


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

__all__ = ["dynamic_metadata", "get_requires_for_dynamic_metadata"]


def __dir__() -> list[str]:
    return __all__


def dynamic_metadata(
    field: str,
    settings: dict[str, object] | None = None,
) -> str:
    # this is a classic implementation, waiting for the release of
    # vcs-versioning and an improved public interface

    if field != "version":
        msg = "Only the 'version' field is supported"
        raise ValueError(msg)

    if settings:
        msg = "No inline configuration is supported"
        raise ValueError(msg)

    from setuptools_scm import Configuration, _get_version

    config = Configuration.from_file("pyproject.toml")
    version: str | None
    try:
        version = _get_version(config, force_write_version_files=True)
    except TypeError:  # setuptools_scm < 8
        version = _get_version(config)

    if version is None:
        msg = (
            f"setuptools-scm was unable to detect version for {config.absolute_root}.\n\n"
            "Make sure you're either building from a fully intact git repository "
            "or PyPI tarballs. Most other sources (such as GitHub's tarballs, a "
            "git checkout without the .git folder) don't contain the necessary "
            "metadata and will not work.\n\n"
            "For example, if you're using pip, instead of "
            "https://github.com/user/proj/archive/master.zip "
            "use git+https://github.com/user/proj.git#egg=proj"
        )

        raise ValueError(msg)

    return version


def get_requires_for_dynamic_metadata(
    _settings: dict[str, object] | None = None,
) -> list[str]:
    return ["setuptools-scm"]



import dataclasses
from pathlib import Path
from typing import List

from .common import Paths

__all__ = ["CMakeFiles", "Input"]


def __dir__() -> List[str]:
    return __all__


@dataclasses.dataclass(frozen=True)
class Input:
    path: Path
    isGenerated: bool = False
    isExternal: bool = False
    isCMake: bool = False


@dataclasses.dataclass(frozen=True)
class CMakeFiles:
    kind: str
    paths: Paths
    inputs: List[Input]



import dataclasses
from pathlib import Path
from typing import List, Optional

from .common import APIVersion, Paths

__all__ = [
    "Archive",
    "Artifact",
    "CodeModel",
    "CommandFragment",
    "Configuration",
    "Dependency",
    "Destination",
    "Directory",
    "Install",
    "Link",
    "Prefix",
    "Project",
    "Source",
    "StringCMakeVersion",
    "Sysroot",
    "Target",
]


def __dir__() -> List[str]:
    return __all__


@dataclasses.dataclass(frozen=True)
class StringCMakeVersion:
    string: str


@dataclasses.dataclass(frozen=True)
class Directory:
    source: Path
    build: Path
    projectIndex: int
    jsonFile: Optional[Path] = None
    parentIndex: Optional[int] = None
    childIndexes: List[int] = dataclasses.field(default_factory=list)
    targetIndexes: List[int] = dataclasses.field(default_factory=list)
    minimumCMakeVersion: Optional[StringCMakeVersion] = None
    hasInstallRule: bool = False


# Directory is currently not resolved automatically.


@dataclasses.dataclass(frozen=True)
class Project:
    name: str
    directoryIndexes: List[int]
    parentIndex: Optional[int] = None
    childIndexes: List[int] = dataclasses.field(default_factory=list)
    targetIndexes: List[int] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True)
class Artifact:
    path: Path


@dataclasses.dataclass(frozen=True)
class Prefix:
    path: Path


@dataclasses.dataclass(frozen=True)
class Destination:
    path: Path
    backtrace: Optional[int] = None


@dataclasses.dataclass(frozen=True)
class Install:
    prefix: Prefix
    destinations: List[Destination]


@dataclasses.dataclass(frozen=True)
class CommandFragment:
    fragment: str
    role: str


@dataclasses.dataclass(frozen=True)
class Sysroot:
    path: Path


@dataclasses.dataclass(frozen=True)
class Link:
    language: str
    commandFragments: Optional[List[CommandFragment]]
    lto: Optional[bool] = None
    sysroot: Optional[Sysroot] = None


@dataclasses.dataclass(frozen=True)
class Archive:
    commandFragments: Optional[List[CommandFragment]] = dataclasses.field(
        default_factory=list
    )
    lto: Optional[bool] = None


@dataclasses.dataclass(frozen=True)
class Dependency:
    id: str
    backtrace: Optional[int] = None


@dataclasses.dataclass(frozen=True)
class Source:
    path: Path
    compileGroupIndex: Optional[int] = None
    sourceGroupIndex: Optional[int] = None
    isGenerated: Optional[bool] = None
    backtrace: Optional[int] = None


@dataclasses.dataclass(frozen=True)
class Target:
    name: str
    id: str
    type: str
    paths: Paths
    sources = List[Source]
    nameOnDisk: Optional[Path] = None
    artifacts: List[Artifact] = dataclasses.field(default_factory=list)
    isGeneratorProvided: Optional[bool] = None
    install: Optional[Install] = None
    link: Optional[Link] = None
    archive: Optional[Archive] = None
    dependencies: List[Dependency] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True)
class Configuration:
    name: str
    projects: List[Project]
    targets: List[Target]
    directories: List[Directory]


@dataclasses.dataclass(frozen=True)
class CodeModel:
    kind: str
    version: APIVersion
    paths: Paths
    configurations: List[Configuration]



import dataclasses
from pathlib import Path
from typing import List, Optional

from .common import APIVersion

__all__ = ["Compiler", "Implicit", "Toolchain", "Toolchains"]


def __dir__() -> List[str]:
    return __all__


@dataclasses.dataclass(frozen=True)
class Implicit:
    includeDirectories: List[Path] = dataclasses.field(default_factory=list)
    linkDirectories: List[Path] = dataclasses.field(default_factory=list)
    linkFrameworkDirectories: List[Path] = dataclasses.field(default_factory=list)
    linkLibraries: List[Path] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True)
class Compiler:
    implicit: Implicit
    path: Optional[Path] = None
    id: Optional[str] = None
    version: Optional[str] = None
    target: Optional[str] = None


@dataclasses.dataclass(frozen=True)
class Toolchain:
    language: str  # Unique, since CMake supports one toolchain per language
    compiler: Compiler
    sourceFileExtensions: List[str] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True)
class Toolchains:
    kind: str = "toolchains"
    version: APIVersion = APIVersion(1, 0)  # noqa: RUF009
    toolchains: List[Toolchain] = dataclasses.field(default_factory=list)



from __future__ import annotations

import contextlib
import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

__all__ = ["process_script_dir"]


def __dir__() -> list[str]:
    return __all__


SHEBANG_PATTERN = re.compile(r"^#!.*(?:python|pythonw|pypy)[0-9.]*([ \t].*)?$")


def process_script_dir(script_dir: Path) -> None:
    for item in script_dir.iterdir():
        content = []
        with contextlib.suppress(UnicodeDecodeError), item.open(encoding="utf-8") as f:
            file_iter = iter(f)
            try:
                # TODO: handle empty files
                first_line = next(file_iter)
            except StopIteration:
                first_line = ""
            match = SHEBANG_PATTERN.match(first_line)
            if match:
                content = [f"#!python{match.group(1) or ''}\n", *file_iter]
        if content:
            with item.open("w", encoding="utf-8") as f:
                f.writelines(content)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from .._compat import tomllib

if TYPE_CHECKING:
    from collections.abc import Mapping

__all__ = [
    "dynamic_metadata",
    "get_requires_for_dynamic_metadata",
]


def __dir__() -> list[str]:
    return __all__


def dynamic_metadata(
    field: str,
    settings: dict[str, list[str] | str],
    project: Mapping[str, Any],
) -> str | dict[str, str]:
    from hatch_fancy_pypi_readme._builder import build_text
    from hatch_fancy_pypi_readme._config import load_and_validate_config

    if field != "readme":
        msg = "Only the 'readme' field is supported"
        raise ValueError(msg)

    if settings:
        msg = "No inline configuration is supported"
        raise ValueError(msg)

    with Path("pyproject.toml").open("rb") as f:
        pyproject_dict = tomllib.load(f)

    config = load_and_validate_config(
        pyproject_dict["tool"]["hatch"]["metadata"]["hooks"]["fancy-pypi-readme"]
    )

    try:
        text = build_text(
            config.fragments,
            getattr(config, "substitutions", []),
            version=project["version"],
            package_name=project["name"],
        )
    except TypeError:
        # hatch-fancy-pypi-readme >=25.1 required for package name
        text = build_text(
            config.fragments,
            getattr(config, "substitutions", []),
            version=project["version"],
        )

    return {
        "content-type": config.content_type,
        "text": text,
    }


def get_requires_for_dynamic_metadata(
    _settings: dict[str, object] | None = None,
) -> list[str]:
    return ["hatch-fancy-pypi-readme>=23.2"]



from __future__ import annotations

import dataclasses
import os
import platform
import shutil
import sys
import sysconfig
import tempfile
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

from packaging.requirements import Requirement
from packaging.tags import Tag
from packaging.utils import canonicalize_name

from .._compat import tomllib
from .._compat.typing import assert_never
from .._logging import LEVEL_VALUE, logger, rich_error, rich_print
from ..builder.builder import Builder, archs_to_tags, get_archs
from ..builder.wheel_tag import WheelTag
from ..cmake import CMake, CMaker
from ..errors import FailedLiveProcessError
from ..format import pyproject_format
from ..settings.skbuild_read_settings import SettingsReader
from ._editable import editable_redirect, libdir_to_installed, mapping_to_modules
from ._init import setup_logging
from ._pathutil import (
    packages_to_file_mapping,
)
from ._scripts import process_script_dir
from ._wheelfile import WheelMetadata, WheelWriter
from .generate import generate_file_contents
from .metadata import get_standard_metadata

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from ..settings.skbuild_model import ScikitBuildSettings

__all__ = ["_build_wheel_impl"]


def __dir__() -> list[str]:
    return __all__


def _make_editable(
    *,
    build_options: Sequence[str] = (),
    install_options: Sequence[str] = (),
    libdir: Path,
    mapping: dict[str, str],
    name: str,
    reload_dir: Path | None,
    settings: ScikitBuildSettings,
    wheel: WheelWriter,
    packages: Iterable[str],
) -> None:
    modules = mapping_to_modules(mapping, libdir)
    installed = libdir_to_installed(libdir)
    if settings.wheel.install_dir.startswith("/"):
        msg = "Editable installs cannot rebuild an absolute wheel.install-dir. Use an override to change if needed."
        raise AssertionError(msg)
    editable_txt = editable_redirect(
        modules=modules,
        installed=installed,
        reload_dir=reload_dir,
        rebuild=settings.editable.rebuild,
        verbose=settings.editable.verbose,
        build_options=build_options,
        install_options=install_options,
        install_dir=settings.wheel.install_dir,
    )

    wheel.writestr(
        f"_{name}_editable.py",
        editable_txt.encode(),
    )
    # Support Cython by adding the source directory directly to the path.
    # This is necessary because Cython does not support sys.meta_path for
    # cimports (as of 3.0.5).
    import_strings = [f"import _{name}_editable", *packages, ""]
    pth_import_paths = "\n".join(import_strings)
    wheel.writestr(
        f"_{name}_editable.pth",
        pth_import_paths.encode(),
    )


def _get_packages(
    *,
    packages: Sequence[str] | Mapping[str, str] | None,
    name: str,
) -> dict[str, str]:
    if packages is not None:
        if isinstance(packages, Mapping):
            return dict(packages)
        return {str(Path(p).name): p for p in packages}

    # Auto package discovery
    packages = {}
    for base_path in (Path("src"), Path("python"), Path()):
        path = base_path / name
        if path.is_dir() and (
            (path / "__init__.py").is_file() or (path / "__init__.pyi").is_file()
        ):
            logger.info("Discovered Python package at {}", path)
            packages[name] = str(path)
            break
    else:
        logger.debug("Didn't find a Python package for {}", name)

    return packages


@dataclasses.dataclass
class WheelImplReturn:
    wheel_filename: str
    settings: ScikitBuildSettings
    mapping: dict[str, str] = dataclasses.field(default_factory=dict)


def _build_wheel_impl(
    wheel_directory: str | None,
    config_settings: dict[str, list[str] | str] | None,
    metadata_directory: str | None,
    *,
    exit_after_config: bool = False,
    editable: bool,
) -> WheelImplReturn:
    """
    Build a wheel or just prepare metadata (if wheel dir is None). Can be editable.
    Handles one retry attempt if "failed" override present.
    """
    state: Literal["sdist", "wheel", "editable", "metadata_wheel", "metadata_editable"]
    if exit_after_config:
        state = "sdist"
    elif wheel_directory is None:
        state = "metadata_editable" if editable else "metadata_wheel"
    else:
        state = "editable" if editable else "wheel"

    pyproject_path = Path("pyproject.toml")
    with pyproject_path.open("rb") as ft:
        pyproject = tomllib.load(ft)

    settings_reader = SettingsReader(
        pyproject, config_settings or {}, state=state, retry=False
    )
    setup_logging(settings_reader.settings.logging.level)

    settings_reader.validate_may_exit()

    if settings_reader.settings.fail:
        if settings_reader.settings.messages.after_failure:
            rich_print(settings_reader.settings.messages.after_failure)
            raise SystemExit(7)
        rich_error("scikit-build-core's fail setting was enabled. Exiting immediately.")

    # Warn if cmake or ninja is in build-system.requires
    requirements = [
        canonicalize_name(Requirement(p).name)
        for p in pyproject.get("build-system", {}).get("requires", [])
    ]
    if "cmake" in requirements:
        logger.warning(
            "cmake should not be in build-system.requires - scikit-build-core will inject it as needed"
        )
    if "ninja" in requirements:
        logger.warning(
            "ninja should not be in build-system.requires - scikit-build-core will inject it as needed"
        )

    try:
        return _build_wheel_impl_impl(
            wheel_directory,
            metadata_directory,
            exit_after_config=exit_after_config,
            editable=editable,
            state=state,
            settings=settings_reader.settings,
            pyproject=pyproject,
        )
    except FailedLiveProcessError as err:
        settings_reader = SettingsReader(
            pyproject, config_settings or {}, state=state, retry=True
        )
        if "failed" not in settings_reader.overrides:
            err.msg = settings_reader.settings.messages.after_failure
            raise

        rich_print(
            "\n***",
            *err.args,
            "- retrying due to override...",
            color="yellow",
        )

        logger.setLevel(LEVEL_VALUE[settings_reader.settings.logging.level])

        settings_reader.validate_may_exit()

        try:
            return _build_wheel_impl_impl(
                wheel_directory,
                metadata_directory,
                exit_after_config=exit_after_config,
                editable=editable,
                state=state,
                settings=settings_reader.settings,
                pyproject=pyproject,
            )
        except FailedLiveProcessError as err2:
            err2.msg = settings_reader.settings.messages.after_failure.format()
            raise


def _build_wheel_impl_impl(
    wheel_directory: str | None,
    metadata_directory: str | None,
    *,
    exit_after_config: bool = False,
    editable: bool,
    state: Literal["sdist", "wheel", "editable", "metadata_wheel", "metadata_editable"],
    settings: ScikitBuildSettings,
    pyproject: dict[str, Any],
) -> WheelImplReturn:
    """
    Build a wheel or just prepare metadata (if wheel dir is None). Can be editable.
    """

    metadata = get_standard_metadata(pyproject, settings)

    if metadata.version is None:
        msg = "project.version is not specified, must be statically present or tool.scikit-build metadata.version.provider configured when dynamic"
        raise AssertionError(msg)

    # Verify PEP 639 replaces license-files
    if metadata.license_files is not None and settings.wheel.license_files:
        msg = "Both project.license-files and tool.scikit-build.wheel.license-files are set, use only one"
        raise AssertionError(msg)

    # Get the closest (normally) importable name
    normalized_name = metadata.name.replace("-", "_").replace(".", "_")

    if settings.wheel.cmake:
        cmake = CMake.default_search(version=settings.cmake.version, env=os.environ)
        cmake_msg = [f"using {{blue}}CMake {cmake.version}{{default}}"]
    else:
        cmake = None
        cmake_msg = []

    if settings.wheel.platlib is None:
        targetlib = "platlib" if settings.wheel.cmake else "purelib"
    else:
        targetlib = "platlib" if settings.wheel.platlib else "purelib"

    rich_print(
        "{green}*** {bold}scikit-build-core {__version__}",
        *cmake_msg,
        f"{{red}}({state})",
    )
    logger.info(
        "Implementation: {} {} on {}",
        sys.implementation.name,
        sys.platform,
        platform.machine(),
    )

    override_wheel_tags = None
    if settings.wheel.tags:
        override_wheel_tags = {Tag(*tag.split("-")) for tag in settings.wheel.tags}

    with tempfile.TemporaryDirectory() as tmpdir:
        build_tmp_folder = Path(tmpdir)
        wheel_dir = build_tmp_folder / "wheel"

        tags = WheelTag.compute_best(
            archs_to_tags(get_archs(os.environ)),
            settings.wheel.py_api,
            expand_macos=settings.wheel.expand_macos_universal_tags,
            root_is_purelib=targetlib == "purelib",
            build_tag=settings.wheel.build_tag,
        )

        # A build dir can be specified, otherwise use a temporary directory
        if cmake is not None and editable and settings.editable.mode == "inplace":
            build_dir = settings.cmake.source_dir
        else:
            build_dir = (
                Path(
                    settings.build_dir.format(
                        **pyproject_format(
                            settings=settings,
                            tags=tags,
                            state=state,
                        )
                    )
                )
                if settings.build_dir
                else build_tmp_folder / "build"
            )
            logger.info("Build directory: {}", build_dir.resolve())

        wheel_dirs = {
            targetlib: wheel_dir / targetlib,
            "data": wheel_dir / "data",
            "headers": wheel_dir / "headers",
            "scripts": wheel_dir / "scripts",
            "null": wheel_dir / "null",
            "metadata": wheel_dir / "metadata",
        }

        for d in wheel_dirs.values():
            d.mkdir(parents=True)

        if ".." in settings.wheel.install_dir:
            msg = "wheel.install_dir must not contain '..'"
            raise AssertionError(msg)
        if settings.wheel.install_dir.startswith("/"):
            if not settings.experimental:
                msg = "Experimental features must be enabled to use absolute paths in wheel.install_dir"
                raise AssertionError(msg)
            if settings.wheel.install_dir[1:].split("/")[0] not in wheel_dirs:
                msg = "Must target a valid wheel directory"
                raise AssertionError(msg)
            install_dir = wheel_dir / settings.wheel.install_dir[1:]
        else:
            install_dir = wheel_dirs[targetlib] / settings.wheel.install_dir

        # Include the metadata license.file entry if provided
        if metadata.license_files:
            license_paths = metadata.license_files
        else:
            license_file_globs = settings.wheel.license_files or [
                "LICEN[CS]E*",
                "COPYING*",
                "NOTICE*",
                "AUTHORS*",
            ]
            if (
                metadata.license
                and not isinstance(metadata.license, str)
                and metadata.license.file
            ):
                license_file_globs.append(str(metadata.license.file))

            license_paths = [
                x for y in license_file_globs for x in Path().glob(y) if x.is_file()
            ]

        for x in license_paths:
            path = wheel_dirs["metadata"] / "licenses" / x
            path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(x, path)

        if (
            settings.wheel.license_files
            and not (wheel_dirs["metadata"] / "licenses").is_dir()
        ):
            logger.warning(
                "No license files found, set wheel.license-files to [] to suppress this warning"
            )

        for gen in settings.generate:
            if gen.location == "source":
                contents = generate_file_contents(gen, metadata)
                gen.path.write_text(contents)
                settings.sdist.include.append(str(gen.path))

        if wheel_directory is None and not exit_after_config:
            if metadata_directory is None:
                msg = "metadata_directory must be specified if wheel_directory is None"
                raise AssertionError(msg)
            wheel = WheelWriter(
                metadata,
                Path(metadata_directory),
                override_wheel_tags or tags.as_tags_set(),
                WheelMetadata(
                    root_is_purelib=targetlib == "purelib",
                    build_tag=settings.wheel.build_tag,
                ),
                wheel_dirs["metadata"],
            )
            dist_info_contents = wheel.dist_info_contents()
            dist_info = Path(metadata_directory) / f"{wheel.name_ver}.dist-info"
            dist_info.mkdir(parents=True)
            for key, data in dist_info_contents.items():
                path = dist_info / key
                if not path.parent.is_dir():
                    path.parent.mkdir(exist_ok=True, parents=True)
                path.write_bytes(data)
            return WheelImplReturn(wheel_filename=dist_info.name, settings=settings)

        for gen in settings.generate:
            contents = generate_file_contents(gen, metadata)
            if gen.location == "source":
                continue
            if gen.location == "build":
                path = build_dir / gen.path
            elif gen.location == "install":
                path = wheel_dirs[targetlib] / gen.path
            else:
                assert_never(gen.location)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(contents, encoding="utf-8")

        build_options = []
        install_options = []

        if cmake is not None:
            config = CMaker(
                cmake,
                source_dir=settings.cmake.source_dir,
                build_dir=build_dir,
                build_type=settings.cmake.build_type,
            )

            builder = Builder(
                settings=settings,
                config=config,
            )

            rich_print("{green}***", "{bold}Configuring CMake...")
            # Setting the install prefix because some libs hardcode CMAKE_INSTALL_PREFIX
            # Otherwise `cmake --install --prefix` would work by itself
            defines = {"CMAKE_INSTALL_PREFIX": install_dir}
            cache_entries: dict[str, str | Path] = {
                f"SKBUILD_{k.upper()}_DIR": v for k, v in wheel_dirs.items()
            }
            cache_entries["SKBUILD_STATE"] = state
            builder.configure(
                defines=defines,
                cache_entries=cache_entries,
                name=metadata.name,
                version=metadata.version,
            )

            if exit_after_config:
                return WheelImplReturn("", settings=settings)

            default_gen = (
                "MSVC"
                if sysconfig.get_platform().startswith("win")
                else "Default Generator"
            )
            generator = builder.get_generator() or default_gen
            rich_print(
                "{green}***",
                f"{{bold}}Building project with {{blue}}{generator}{{default}}...",
            )
            build_args: list[str] = []
            builder.build(build_args=build_args)

            if not (editable and settings.editable.mode == "inplace"):
                rich_print(
                    "{green}***",
                    "{bold}Installing project into wheel...",
                )
                builder.install(install_dir)

            if not builder.config.single_config and builder.config.build_type:
                build_options += ["--config", builder.config.build_type]
                install_options += ["--config", builder.config.build_type]
            if builder.settings.cmake.verbose:
                build_options.append("-v")

        assert wheel_directory is not None

        rich_print("{green}***", f"{{bold}}Making {state}...")
        packages = _get_packages(
            packages=settings.wheel.packages,
            name=normalized_name,
        )
        assert settings.sdist.inclusion_mode is not None
        mapping = packages_to_file_mapping(
            packages=packages,
            platlib_dir=wheel_dirs[targetlib],
            include=settings.sdist.include,
            src_exclude=settings.sdist.exclude,
            target_exclude=settings.wheel.exclude,
            build_dir=settings.build_dir,
            mode=settings.sdist.inclusion_mode,
        )

        if not editable:
            for filepath, package_dir in mapping.items():
                Path(package_dir).parent.mkdir(exist_ok=True, parents=True)
                shutil.copy2(filepath, package_dir)

            process_script_dir(wheel_dirs["scripts"])

        with WheelWriter(
            metadata,
            Path(wheel_directory),
            override_wheel_tags or tags.as_tags_set(),
            WheelMetadata(
                root_is_purelib=targetlib == "purelib",
                build_tag=settings.wheel.build_tag,
            ),
            wheel_dirs["metadata"],
        ) as wheel:
            wheel.build(wheel_dirs, exclude=settings.wheel.exclude)

            str_pkgs = (
                str(Path.cwd().joinpath(p).parent.resolve()) for p in packages.values()
            )
            if editable and settings.editable.mode == "redirect":
                reload_dir = build_dir.resolve() if settings.build_dir else None

                _make_editable(
                    build_options=build_options,
                    install_options=install_options,
                    libdir=wheel_dirs[targetlib],
                    mapping=mapping,
                    reload_dir=reload_dir,
                    settings=settings,
                    wheel=wheel,
                    name=normalized_name,
                    packages=str_pkgs,
                )
            elif editable and settings.editable.mode == "inplace":
                if not packages:
                    msg = "Editable inplace mode requires at least one package"
                    raise AssertionError(msg)

                wheel.writestr(
                    f"_{normalized_name}_editable.pth",
                    "\n".join(str_pkgs).encode(),
                )

    if metadata_directory is not None:
        dist_info_contents = wheel.dist_info_contents()
        dist_info = Path(metadata_directory)
        for key, data in dist_info_contents.items():
            path = dist_info / key
            previous_data = path.read_bytes()
            if previous_data != data:
                msg = f"Metadata mismatch in {key}"
                logger.error("{}: {!r} != {!r}", msg, previous_data, data)
                raise AssertionError(msg)

    wheel_filename: str = wheel.wheelpath.name
    rich_print("{green}***", f"{{bold}}Created{{normal}} {wheel_filename}")
    if settings.messages.after_success:
        rich_print(settings.messages.after_success)
    return WheelImplReturn(
        wheel_filename=wheel_filename, mapping=mapping, settings=settings
    )



import argparse
import json
from pathlib import Path
from typing import Literal

from .._compat import tomllib
from .._logging import rich_warning
from ..builder._load_provider import process_dynamic_metadata
from . import (
    get_requires_for_build_editable,
    get_requires_for_build_sdist,
    get_requires_for_build_wheel,
)


def main_project_table(_args: argparse.Namespace, /) -> None:
    """Get the full project table, including dynamic metadata."""
    with Path("pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)

    project = pyproject.get("project", {})
    metadata = pyproject.get("tool", {}).get("scikit-build", {}).get("metadata", {})
    new_project = process_dynamic_metadata(project, metadata)
    print(json.dumps(new_project, indent=2))


def main_requires(args: argparse.Namespace, /) -> None:
    get_requires(args.mode)


def get_requires(mode: Literal["sdist", "wheel", "editable"]) -> None:
    """Get the build requirements."""

    with Path("pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)

    requires = pyproject.get("build-system", {}).get("requires", [])
    backend = pyproject.get("build-system", {}).get("build-backend", "")
    if backend != "scikit_build_core.build":
        rich_warning("Might not be a scikit-build-core project.")
    if mode == "sdist":
        requires += get_requires_for_build_sdist({})
    elif mode == "wheel":
        requires += get_requires_for_build_wheel({})
    elif mode == "editable":
        requires += get_requires_for_build_editable({})
    print(json.dumps(sorted(set(requires)), indent=2))


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="python -m scikit_build_core.build",
        allow_abbrev=False,
        description="Build backend utilities.",
    )

    subparsers = parser.add_subparsers(required=True, help="Commands")
    requires = subparsers.add_parser(
        "requires",
        help="Get the build requirements",
        description="Includes the static build requirements, the dynamically generated ones, and dynamic-metadata ones.",
    )
    requires.set_defaults(func=main_requires)
    requires.add_argument(
        "--mode",
        choices=["sdist", "wheel", "editable"],
        default="wheel",
        help="The build mode to get the requirements for",
    )

    project_table = subparsers.add_parser(
        "project-table",
        help="Get the full project table, including dynamic metadata",
        description="Processes static and dynamic metadata without triggering the backend, only handles scikit-build-core's dynamic metadata.",
    )
    project_table.set_defaults(func=main_project_table)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



from __future__ import annotations

import functools
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any

from . import _process_dynamic_metadata

if TYPE_CHECKING:
    from collections.abc import Mapping

__all__ = ["dynamic_metadata"]


def __dir__() -> list[str]:
    return __all__


KEYS = {"input", "regex", "result", "remove"}


def _process(match: re.Match[str], remove: str, result: str) -> str:
    retval = result.format(*match.groups(), **match.groupdict())
    if remove:
        retval = re.sub(remove, "", retval)
    return retval


def dynamic_metadata(
    field: str,
    settings: Mapping[str, Any],
) -> str:
    # Input validation
    if settings.keys() > KEYS:
        msg = f"Only {KEYS} settings allowed by this plugin"
        raise RuntimeError(msg)
    if "input" not in settings:
        msg = "Must contain the 'input' setting to perform a regex on"
        raise RuntimeError(msg)
    if field != "version" and "regex" not in settings:
        msg = "Must contain the 'regex' setting if not getting version"
        raise RuntimeError(msg)
    for key in KEYS:
        if key in settings and not isinstance(settings[key], str):
            msg = f"Setting {key!r} must be a string"
            raise RuntimeError(msg)

    input_filename = settings["input"]
    regex = settings.get(
        "regex",
        r'(?i)^(__version__|VERSION)(?: ?\: ?str)? *= *([\'"])v?(?P<value>.+?)\2',
    )
    result = settings.get("result", "{value}")
    assert isinstance(result, str)
    remove = settings.get("remove", "")

    with Path(input_filename).open(encoding="utf-8") as f:
        match = re.search(regex, f.read(), re.MULTILINE)

    if not match:
        msg = f"Couldn't find {regex!r} in {input_filename}"
        raise RuntimeError(msg)

    return _process_dynamic_metadata(
        field, functools.partial(_process, match, remove), result
    )



version: str
# pylint: disable-next=unsupported-binary-operation
version_tuple: tuple[int, int, int, str, str] | tuple[int, int, int]

# Note that newer versions of setuptools_scm also add __version__, but we are
# not forcing new versions of setuptools_scm, so only these imports are allowed.



from __future__ import annotations

import dataclasses
import os
import re
import shlex
import sys
import sysconfig
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .. import __version__
from .._compat.importlib import metadata, resources
from .._logging import logger
from ..resources import find_python
from .generator import set_environment_for_gen
from .sysconfig import (
    get_numpy_include_dir,
    get_platform,
    get_python_include_dir,
    get_python_library,
    get_soabi,
)

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable, Mapping, Sequence

    from packaging.version import Version

    from ..cmake import CMaker
    from ..settings.skbuild_model import ScikitBuildSettings

__all__ = ["Builder", "archs_to_tags", "get_archs"]

DIR = Path(__file__).parent.resolve()


def __dir__() -> list[str]:
    return __all__


# TODO: cross-compile support for other platforms
def get_archs(env: Mapping[str, str], cmake_args: Sequence[str] = ()) -> list[str]:
    """
    Takes macOS platform settings and returns a list of platforms.

    Example (macOS):
        ARCHFLAGS="-arch x86_64" -> ["x86_64"]
        ARCHFLAGS="-arch x86_64 -arch arm64" -> ["x86_64", "arm64"]

    Returns an empty list otherwise or if ARCHFLAGS is not set.
    """

    if sys.platform.startswith("darwin"):
        for cmake_arg in cmake_args:
            if "CMAKE_SYSTEM_PROCESSOR" in cmake_arg:
                return [cmake_arg.split("=")[1]]
        return re.findall(r"-arch (\S+)", env.get("ARCHFLAGS", ""))
    if sys.platform.startswith("win") and get_platform(env) == "win-arm64":
        return ["win_arm64"]

    return []


def archs_to_tags(archs: list[str]) -> list[str]:
    """
    Convert a list of architectures to a list of tags (e.g. "universal2").
    """
    if sys.platform.startswith("darwin") and set(archs) == {"arm64", "x86_64"}:
        return ["universal2"]
    return archs


def _filter_env_cmake_args(env_cmake_args: list[str]) -> Generator[str, None, None]:
    """
    Filter out CMake arguments that are not supported from CMAKE_ARGS.
    """

    unsupported_args = ("-DCMAKE_BUILD_TYPE", "-DCMAKE_INSTALL_PREFIX")

    for arg in env_cmake_args:
        if arg.startswith(unsupported_args):
            logger.warning("Unsupported CMAKE_ARGS ignored: {}", arg)
        else:
            yield arg


def _sanitize_path(path: Any) -> list[Path]:
    # This handles classes like:
    # MultiplexedPath from importlib.resources.readers (3.11+)
    # MultiplexedPath from importlib.readers (3.10)
    # MultiplexedPath from importlib_resources.readers
    if hasattr(path, "_paths"):
        # pylint: disable-next=protected-access
        return [Path(os.fspath(p)) for p in path._paths]

    return [Path(os.fspath(path))]


@dataclasses.dataclass
class Builder:
    settings: ScikitBuildSettings
    config: CMaker

    def get_cmake_args(self) -> list[str]:
        """
        Get CMake args from the settings and environment.
        """
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSX on conda-forge)
        env_cmake_args: list[str] = list(
            filter(None, shlex.split(self.config.env.get("CMAKE_ARGS", "")))
        )

        if env_cmake_args:
            logger.debug("Env CMAKE_ARGS: {}", env_cmake_args)

        return [*self.settings.cmake.args, *_filter_env_cmake_args(env_cmake_args)]

    def get_generator(self, *args: str) -> str | None:
        return self.config.get_generator(
            *self.get_cmake_args(), *args, defines=self.settings.cmake.define
        )

    def _get_entry_point_search_path(self, entry_point: str) -> dict[str, list[Path]]:
        """Get the search path dict from the entry points"""
        search_paths = {}
        eps = metadata.entry_points(group=entry_point)
        if eps:
            logger.debug(
                "Loading search paths {} from entry-points: {}", entry_point, len(eps)
            )
            for ep in eps:
                ep_value = _sanitize_path(resources.files(ep.load()))
                logger.debug("{}: {} -> {}", ep.name, ep.value, ep_value)
                if ep_value:
                    search_paths[ep.name] = ep_value
        return search_paths

    def configure(
        self,
        *,
        defines: Mapping[str, str | bool | Path],
        cache_entries: Mapping[str, str | Path] | None = None,
        name: str | None = None,
        version: Version | None = None,
        limited_api: bool | None = None,
        configure_args: Iterable[str] = (),
    ) -> None:
        cmake_defines = {
            k: ("TRUE" if v else "FALSE") if isinstance(v, bool) else str(v)
            for k, v in defines.items()
        }

        # Add any extra CMake modules
        self.config.module_dirs.extend(
            p
            for ep_paths in self._get_entry_point_search_path("cmake.module").values()
            for p in ep_paths
        )
        logger.debug("cmake.modules: {}", self.config.module_dirs)

        # Add any extra CMake prefixes
        self.config.prefix_dirs.extend(
            p
            for ep_paths in self._get_entry_point_search_path("cmake.prefix").values()
            for p in ep_paths
        )
        logger.debug("cmake.prefix: {}", self.config.prefix_dirs)

        # Add all CMake roots
        # TODO: Check for unique uppercase names
        self.config.prefix_roots.update(self._get_entry_point_search_path("cmake.root"))
        logger.debug("cmake.root: {}", self.config.prefix_roots)

        # Add site-packages to the prefix path for CMake
        site_packages = Path(sysconfig.get_path("purelib"))
        if self.settings.search.site_packages:
            self.config.prefix_dirs.append(site_packages)
            logger.debug("SITE_PACKAGES: {}", site_packages)
            if site_packages != DIR.parent.parent:
                self.config.prefix_dirs.append(DIR.parent.parent)
                logger.debug("Extra SITE_PACKAGES: {}", DIR.parent.parent)
                logger.debug("PATH: {}", sys.path)

        # Add the FindPython backport if needed
        if self.config.cmake.version < self.settings.backport.find_python:
            fp_dir = Path(find_python.__file__).parent.resolve()
            self.config.module_dirs.append(fp_dir)
            logger.debug("FindPython backport activated at {}", fp_dir)

        current_gen = self.get_generator(*configure_args)
        local_def = set_environment_for_gen(
            current_gen, self.config.cmake, self.config.env, self.settings.ninja
        )
        cmake_defines.update(local_def)

        cache_config: dict[str, str | Path | bool] = {
            "SKBUILD": "2",
            "SKBUILD_CORE_VERSION": __version__,
        }

        if name is not None:
            canonical_name = name.replace("-", "_").replace(".", "_")
            cache_config["SKBUILD_PROJECT_NAME"] = canonical_name
        if version is not None:
            cache_config["SKBUILD_PROJECT_VERSION"] = ".".join(
                str(v) for v in version.release
            )
            cache_config["SKBUILD_PROJECT_VERSION_FULL"] = str(version)

        if limited_api is None:
            if self.settings.wheel.py_api.startswith("cp3"):
                target_minor_version = int(self.settings.wheel.py_api[3:])
                limited_api = target_minor_version <= sys.version_info.minor
            else:
                limited_api = False

        if limited_api and sys.implementation.name != "cpython":
            limited_api = False
            logger.info("PyPy doesn't support the Limited API, ignoring")

        if limited_api and sysconfig.get_config_var("Py_GIL_DISABLED"):
            limited_api = False
            logger.info(
                "Free-threaded Python doesn't support the Limited API currently, ignoring"
            )

        python_library = get_python_library(self.config.env, abi3=False)
        python_sabi_library = (
            get_python_library(self.config.env, abi3=True) if limited_api else None
        )
        python_include_dir = get_python_include_dir()
        numpy_include_dir = get_numpy_include_dir()

        # Warning for CPython 3.13.4 Windows bug
        if (
            sys.implementation.name == "CPython"
            and sys.version_info[:3] == (3, 13, 4)
            and sys.platform.startswith("win32")
            and not sysconfig.get_config_var("Py_GIL_DISABLED")
        ):  # pragma: nocover
            logger.warning(
                "Python 3.13.4 on Windows is broken for building, 3.13.5 was rushed out to fix it. Use an older, newer, or free-threaded version instead."
            )

        if self.settings.cmake.python_hints:
            # Classic Find Python
            cache_config["PYTHON_EXECUTABLE"] = Path(sys.executable)
            cache_config["PYTHON_INCLUDE_DIR"] = python_include_dir
            if python_library:
                cache_config["PYTHON_LIBRARY"] = python_library

            # Modern Find Python
            for prefix in ("Python", "Python3"):
                cache_config[f"{prefix}_EXECUTABLE"] = Path(sys.executable)
                cache_config[f"{prefix}_ROOT_DIR"] = Path(sys.base_exec_prefix)
                cache_config[f"{prefix}_INCLUDE_DIR"] = python_include_dir
                cache_config[f"{prefix}_FIND_REGISTRY"] = "NEVER"
                # FindPython may break if this is set - only useful on Windows
                if python_library and sysconfig.get_platform().startswith("win"):
                    cache_config[f"{prefix}_LIBRARY"] = python_library
                if python_sabi_library and sysconfig.get_platform().startswith("win"):
                    cache_config[f"{prefix}_SABI_LIBRARY"] = python_sabi_library
                if numpy_include_dir:
                    cache_config[f"{prefix}_NumPy_INCLUDE_DIR"] = numpy_include_dir

        cache_config["SKBUILD_SOABI"] = get_soabi(self.config.env, abi3=limited_api)

        # Allow CMakeLists to detect this is supposed to be a limited ABI build
        cache_config["SKBUILD_SABI_COMPONENT"] = (
            "Development.SABIModule" if limited_api else ""
        )

        # Allow users to detect the version requested in settings
        py_api = self.settings.wheel.py_api
        cache_config["SKBUILD_SABI_VERSION"] = (
            f"{py_api[2]}.{py_api[3:]}"
            if limited_api and py_api.startswith("cp")
            else ""
        )

        if cache_entries:
            cache_config.update(cache_entries)

        self.config.init_cache(cache_config)

        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = get_archs(self.config.env)
            if archs:
                cmake_defines["CMAKE_OSX_ARCHITECTURES"] = ";".join(archs)

        # Add the pre-defined or passed CMake defines
        cmake_defines.update(self.settings.cmake.define)

        self.config.configure(
            defines=cmake_defines,
            cmake_args=[*self.get_cmake_args(), *configure_args],
            toolchain=self.settings.cmake.toolchain_file,
        )

    def build(self, build_args: Sequence[str]) -> None:
        build_tool_args = self.settings.build.tool_args
        if build_tool_args:
            build_args = [*build_args, "--", *build_tool_args]

        self.config.build(
            build_args=build_args,
            targets=self.settings.build.targets,
            verbose=self.settings.build.verbose,
        )

    def install(self, install_dir: Path | None) -> None:
        """
        Install to a path.

        Warning: if a package hard-codes CMAKE_INSTALL_PREFIX in the install
        commands, this will not rewrite those; set that variable when
        configuring for maximum compatibility.
        """
        components = self.settings.install.components
        strip = self.settings.install.strip
        assert strip is not None
        self.config.install(install_dir, strip=strip, components=components)



from __future__ import annotations

import ast
import dataclasses
import inspect
import textwrap
import typing
from pathlib import Path

from packaging.specifiers import SpecifierSet
from packaging.version import Version

from .. import __version__
from .._compat.typing import Annotated, get_args, get_origin

if typing.TYPE_CHECKING:
    from collections.abc import Generator


__all__ = ["pull_docs"]


def __dir__() -> list[str]:
    return __all__


version_display = ".".join(__version__.split(".")[:2])


def _get_value(value: ast.expr) -> str:
    assert isinstance(value, ast.Constant)
    assert isinstance(value.value, str)
    return value.value


def pull_docs(dc: type[object]) -> dict[str, str]:
    """
    Pulls documentation from a dataclass.
    """
    t = ast.parse(inspect.getsource(dc))
    (obody,) = t.body
    assert isinstance(obody, ast.ClassDef)
    body = obody.body
    return {
        assign.target.id: textwrap.dedent(_get_value(expr.value)).strip()  # type: ignore[union-attr]
        for assign, expr in zip(body[:-1], body[1:])
        if isinstance(assign, ast.AnnAssign) and isinstance(expr, ast.Expr)
    }


@dataclasses.dataclass(frozen=True)
class DCDoc:
    name: str
    type: str
    default: str
    docs: str
    field: dataclasses.Field[typing.Any]
    deprecated: bool = False
    override_only: bool = False


def sanitize_default_field(text: str) -> str:
    return text.replace("'", '"').replace("True", "true").replace("False", "false")


def is_optional(field: type) -> bool:
    return get_origin(field) is typing.Union and type(None) in get_args(field)


def get_display_type(field_type: type | str) -> str:
    if isinstance(field_type, str):
        return field_type
    if is_optional(field_type):
        # Special case for optional, we just take the first part
        return get_display_type(get_args(field_type)[0])
    # Handle built-ins
    if get_origin(field_type) is dict:
        key_display = get_display_type(get_args(field_type)[0])
        val_display = get_display_type(get_args(field_type)[1])
        return f"dict[{key_display},{val_display}]"
    if get_origin(field_type) is list:
        return f"list[{get_display_type(get_args(field_type)[0])}]"
    # Handle other typing specials
    if get_origin(field_type) is typing.Literal:
        return " | ".join(f'"{x}"' for x in get_args(field_type))
    if get_origin(field_type) is Annotated:
        # For annotated assume we always want the second item
        return get_display_type(get_args(field_type)[1])
    if field_type is typing.Any:
        # Workaround for python<3.10 where typing.Any.__name__ does not evaluate
        return "Any"
    # Otherwise just get the formatted form of the `type` object
    return field_type.__name__


def mk_docs(dc: type[object], prefix: str = "") -> Generator[DCDoc, None, None]:
    """
    Makes documentation for a dataclass.
    """
    assert dataclasses.is_dataclass(dc)
    docs = pull_docs(dc)

    for field in dataclasses.fields(dc):
        field_type = field.type
        if dataclasses.is_dataclass(field_type) and isinstance(field_type, type):
            yield from mk_docs(field_type, prefix=f"{prefix}{field.name}.")
            continue

        if get_origin(field.type) is list:
            field_type = get_args(field.type)[0]
            if dataclasses.is_dataclass(field_type) and isinstance(field_type, type):
                yield from mk_docs(field_type, prefix=f"{prefix}{field.name}[].")
                continue

        if default_before_format := field.metadata.get("display_default", None):
            assert isinstance(default_before_format, str)
            default = default_before_format.format(
                version=version_display,
            )
        elif field.default is not dataclasses.MISSING and field.default is not None:
            default = repr(
                str(field.default)
                if isinstance(field.default, (Path, Version, SpecifierSet))
                else field.default
            )
        elif field.default_factory is not dataclasses.MISSING:
            default = repr(field.default_factory())
        else:
            default = '""'

        yield DCDoc(
            name=f"{prefix}{field.name}".replace("_", "-"),
            type=field.metadata.get("display_type", get_display_type(field.type)),
            default=sanitize_default_field(default),
            docs=docs[field.name],
            field=field,
            deprecated=field.metadata.get("deprecated", False),
            override_only=field.metadata.get("override_only", False),
        )



from __future__ import annotations

import copy
import json
from typing import Any

from ..resources import resources

__all__ = ["generate_skbuild_schema", "get_skbuild_schema"]


def __dir__() -> list[str]:
    return __all__


METADATA = [
    "version",
    "description",
    "license",
    "readme",
    "requires-python",
    "dependencies",
    "optional-dependencies",
    "entrypoints",
    "authors",
    "maintainers",
    "urls",
    "classifiers",
    "keywords",
    "scripts",
    "gui-scripts",
]


def generate_skbuild_schema(tool_name: str = "scikit-build") -> dict[str, Any]:
    "Generate the complete schema for scikit-build settings."
    assert tool_name == "scikit-build", "Only scikit-build is supported."

    from .json_schema import to_json_schema
    from .skbuild_model import ScikitBuildSettings

    schema = {
        "$schema": "http://json-schema.org/draft-07/schema",
        "$id": "https://github.com/scikit-build/scikit-build-core/blob/main/src/scikit_build_core/resources/scikit-build.schema.json",
        "description": "Scikit-build-core's settings.",
        **to_json_schema(ScikitBuildSettings, normalize_keys=True),
    }

    # Manipulate a bit to get better validation
    # This is making the generate's template or template-path required
    generate = schema["properties"]["generate"]["items"]
    for prop in generate["properties"].values():
        if prop.get("type", "") == "string":
            prop["minLength"] = 1
    generate_tmpl = copy.deepcopy(generate)
    generate_path = copy.deepcopy(generate)

    generate_tmpl["required"] = ["path", "template"]
    del generate_tmpl["properties"]["template-path"]
    del generate_tmpl["properties"]["template"]["default"]

    generate_path["required"] = ["path", "template-path"]
    del generate_path["properties"]["template"]

    schema["properties"]["generate"]["items"] = {
        "oneOf": [generate_tmpl, generate_path]
    }

    schema["$defs"] = {
        "metadata": {
            "type": "object",
            "properties": {
                "provider": {"type": "string"},
                "provider-path": {"type": "string"},
            },
        }
    }

    del schema["properties"]["metadata"]["patternProperties"]
    schema["properties"]["metadata"]["additionalProperties"] = False
    schema["properties"]["metadata"]["properties"] = {
        m: {"$ref": "#/$defs/metadata"} for m in METADATA
    }

    inherit_only_props = {
        k: v
        for k, v in schema["properties"].items()
        if v.get("scikit-build:override-only")
    }
    schema["properties"] = {
        k: v
        for k, v in schema["properties"].items()
        if not v.get("scikit-build:override-only")
    }
    for v in inherit_only_props.values():
        del v["scikit-build:override-only"]
    props = {k: {"$ref": f"#/properties/{k}"} for k in schema["properties"]}
    props.update(inherit_only_props)

    schema["$defs"]["if_overrides"] = {
        "type": "object",
        "minProperties": 1,
        "additionalProperties": False,
        "properties": {
            "scikit-build-version": {
                "type": "string",
                "description": "The version of scikit-build-version. Takes a specifier set.",
            },
            "python-version": {
                "type": "string",
                "description": "The two-digit Python version. Takes a specifier set.",
            },
            "implementation-name": {
                "type": "string",
                "description": "The value of `sys.implementation.name`. Takes a regex",
            },
            "implementation-version": {
                "type": "string",
                "description": "Derived from `sys.implementation.version`, following PEP 508. Takes a specifier set.",
            },
            "platform-system": {
                "type": "string",
                "description": "The value of `sys.platform`. Takes a regex.",
            },
            "platform-machine": {
                "type": "string",
                "description": "The value of `platform.machine()`. Takes a regex.",
            },
            "platform-node": {
                "type": "string",
                "description": "The value of `platform.node()`. Takes a regex.",
            },
            "state": {
                "type": "string",
                "description": "The state of the build, one of `sdist`, `wheel`, `editable`, `metadata_wheel`, and `metadata_editable`. Takes a regex.",
            },
            "from-sdist": {
                "type": "boolean",
                "description": "Whether the build is from an sdist.",
            },
            "failed": {
                "type": "boolean",
                "description": "Matches if the build fails. A build will be retried if there is at least one matching override with this set to true.",
            },
            "system-cmake": {
                "type": "string",
                "description": "The version of CMake found on the system. Takes a specifier set.",
            },
            "cmake-wheel": {
                "type": "boolean",
                "description": "Whether a cmake wheel is known to be provided for this system.",
            },
            "abi-flags": {
                "type": "string",
                "description": "A sorted string of the abi flags. Takes a regex.",
            },
            "env": {
                "type": "object",
                "patternProperties": {
                    ".*": {"oneOf": [{"type": "string"}, {"type": "boolean"}]}
                },
                "additionalProperties": False,
                "minProperties": 1,
                "description": "A table of environment variables mapped to either string regexs, or booleans. Valid 'truthy' environment variables are case insensitive `true`, `on`, `yes`, `y`, `t`, or a number more than 0.",
            },
        },
    }
    schema["$defs"]["inherit"] = {
        "enum": ["none", "append", "prepend"],
        "default": "none",
    }

    inherit_props: dict[str, Any] = {
        k: {
            kk: {"$ref": "#/$defs/inherit"}
            for kk, vv in v["properties"].items()
            if vv.get("type", "") in {"object", "array"}
            or any(
                vvv.get("type", "") in {"object", "array"}
                for vvv in vv.get("oneOf", {})
            )
        }
        for k, v in schema["properties"].items()
        if v.get("type", "") == "object"
    }
    inherit_props = {
        k: {"type": "object", "additionalProperties": False, "properties": v}
        for k, v in inherit_props.items()
        if v
    }

    schema["properties"]["overrides"] = {
        "type": "array",
        "description": "A list of overrides to apply to the settings, based on the `if` selector.",
        "items": {
            "type": "object",
            "required": ["if"],
            "minProperties": 2,
            "additionalProperties": False,
            "properties": {
                "if": {
                    "anyOf": [
                        {"$ref": "#/$defs/if_overrides"},
                        {
                            "type": "object",
                            "properties": {"any": {"$ref": "#/$defs/if_overrides"}},
                            "required": ["any"],
                            "additionalProperties": False,
                        },
                    ]
                },
                "inherit": {
                    "type": "object",
                    "properties": inherit_props,
                    "additionalProperties": False,
                },
                **props,
            },
        },
    }

    return schema


def get_skbuild_schema(tool_name: str = "scikit-build") -> dict[str, Any]:
    "Get the stored complete schema for scikit-build settings."
    assert tool_name == "scikit-build", "Only scikit-build is supported."

    with resources.joinpath("scikit-build.schema.json").open(encoding="utf-8") as f:
        return json.load(f)  # type: ignore[no-any-return]


if __name__ == "__main__":
    d = generate_skbuild_schema()
    print(json.dumps(d, indent=2))  # noqa: T201



from __future__ import annotations

import sys
from pathlib import Path

__all__ = ["stateless_query"]


def __dir__() -> list[str]:
    return __all__


def stateless_query(build_dir: Path) -> Path:
    api_dir = build_dir / ".cmake/api/v1"
    query = api_dir.joinpath("query")
    query.mkdir(parents=True, exist_ok=True)

    query.joinpath("codemodel-v2").touch()
    query.joinpath("cache-v2").touch()
    query.joinpath("cmakeFiles-v1").touch()
    query.joinpath("toolchains-v1").touch()
    return api_dir / "reply"


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        prog="python -m scikit_build_core.file_api.query",
        allow_abbrev=False,
        description="Write a stateless query to a build directory",
    )
    parser.add_argument("build_dir", type=Path, help="Path to the build directory")
    args = parser.parse_args()

    result = stateless_query(args.build_dir)
    sys.stdout.write(f"{result}\n")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Mapping

from . import _process_dynamic_metadata

__all__ = ["dynamic_metadata"]


def __dir__() -> list[str]:
    return __all__


KEYS = {"result"}


def dynamic_metadata(
    field: str,
    settings: Mapping[str, str | list[str] | dict[str, str] | dict[str, list[str]]],
    project: Mapping[str, Any],
) -> str | list[str] | dict[str, str] | dict[str, list[str]]:
    if settings.keys() > KEYS:
        msg = f"Only {KEYS} settings allowed by this plugin"
        raise RuntimeError(msg)

    if "result" not in settings:
        msg = "Must contain the 'result' setting with a template substitution"
        raise RuntimeError(msg)

    result = settings["result"]

    return _process_dynamic_metadata(
        field,
        lambda r: r.format(project=project),
        result,
    )



from __future__ import annotations

import dataclasses
import textwrap
import typing

from .documentation import mk_docs
from .skbuild_model import ScikitBuildSettings

if typing.TYPE_CHECKING:
    from .documentation import DCDoc

__all__ = ["mk_skbuild_docs"]


def __dir__() -> list[str]:
    return __all__


@dataclasses.dataclass
class Document:
    items: list[Item]

    def format(self) -> str:
        return "\n".join(item.format() for item in self.items)


@dataclasses.dataclass
class Item:
    TEMPLATE: typing.ClassVar[str] = textwrap.dedent("""\
    {summary}
    {item.name} = {item.default}
    """)
    item: DCDoc

    def format(self) -> str:
        summary = self.item.docs.split("\n", maxsplit=1)[0]
        return self.TEMPLATE.format(
            item=self.item,
            summary=f"# {summary}",
        )


def mk_skbuild_docs() -> str:
    """
    Makes documentation for the skbuild model.
    """
    doc = Document(
        [
            Item(item)
            for item in mk_docs(ScikitBuildSettings)
            if not item.deprecated and not item.override_only
        ]
    )
    return doc.format()


if __name__ == "__main__":
    print(mk_skbuild_docs())  # noqa: T201



from __future__ import annotations

from ._logging import rich_print

__all__ = ["main"]


def __dir__() -> list[str]:
    return __all__


def main() -> None:
    rich_print("{blue}A top level CLI is not currently provided for scikit-build-core.")
    rich_print("{blue}However, the following modules have CLI utilities:")
    rich_print(
        "  python -m scikit_build_core.build requires       {green}Get the build requirements"
    )
    rich_print(
        "  python -m scikit_build_core.build project-table  {green}Get the project table (with dynamic metadata)"
    )
    rich_print(
        "  python -m scikit_build_core.builder              {green}Info about the system"
    )
    rich_print(
        "  python -m scikit_build_core.builder.wheel_tag    {green}Info about the computed wheel tag"
    )
    rich_print(
        "  python -m scikit_build_core.builder.sysconfig    {green}Info from sysconfig"
    )
    rich_print(
        "  python -m scikit_build_core.file_api.query       {green}Request CMake file API"
    )
    rich_print(
        "  python -m scikit_build_core.file_api.reply       {green}Process CMake file API"
    )
    rich_print()


if __name__ == "__main__":
    main()



import dataclasses
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, TypedDict, Union

from packaging.specifiers import SpecifierSet
from packaging.version import Version

from .._compat.typing import Annotated, Self

__all__ = [
    "BackportSettings",
    "BuildSettings",
    "CMakeSettings",
    "CMakeSettingsDefine",
    "EditableSettings",
    "GenerateSettings",
    "InstallSettings",
    "LoggingSettings",
    "MessagesSettings",
    "NinjaSettings",
    "SDistSettings",
    "ScikitBuildSettings",
    "SearchSettings",
    "SettingsFieldMetadata",
    "WheelSettings",
]


def __dir__() -> List[str]:
    return __all__


class SettingsFieldMetadata(TypedDict, total=False):
    display_default: Optional[str]
    deprecated: bool
    override_only: bool
    """Do not allow the field to be a top-level table."""


class CMakeSettingsDefine(str):
    """
    A str subtype for automatically normalizing bool and list values
    to the CMake representation in the `cmake.define` settings key.
    """

    __slots__ = ()

    json_schema = Union[str, bool, List[str]]

    def __new__(cls, raw: Union[str, bool, List[str]]) -> Self:
        def escape_semicolons(item: str) -> str:
            return item.replace(";", r"\;")

        if isinstance(raw, bool):
            value = "TRUE" if raw else "FALSE"
        elif isinstance(raw, list):
            value = ";".join(map(escape_semicolons, raw))
        else:
            value = raw

        return super().__new__(cls, value)


@dataclasses.dataclass
class CMakeSettings:
    minimum_version: Optional[Version] = dataclasses.field(
        default=None, metadata=SettingsFieldMetadata(deprecated=True)
    )
    """
    DEPRECATED in 0.8; use version instead.
    """

    version: Optional[SpecifierSet] = None
    """
    The versions of CMake to allow as a python-compatible specifier.

    If CMake is not present on the system or does not pass this specifier, it will
    be downloaded via PyPI if possible with the equivalent specifier used.

    An empty string will disable this check.

    Special cases:
      - On scikit-build-core 0.10+ ``CMakeLists.txt`` is the default value otherwise it's
        ``>=3.15``.
      - If ``CMakeLists.txt`` is passed, the ``cmake_minimum_required`` is read from the
        CMakeLists.txt file, using that as the minimum specifier. If the file fails to read,
        ``>=3.15`` is used instead.

    .. seealso::
       :confval:`ninja.version`
    """

    args: List[str] = dataclasses.field(default_factory=list)
    """
    A list of args to pass to CMake when configuring the project.

    Setting this in config or envvar will override toml.

    .. seealso::
       :confval:`cmake.define`
    """

    define: Annotated[Dict[str, CMakeSettingsDefine], "EnvVar"] = dataclasses.field(
        default_factory=dict
    )
    """
    A table of defines to pass to CMake when configuring the project. Additive.
    """

    verbose: Optional[bool] = dataclasses.field(
        default=None, metadata=SettingsFieldMetadata(deprecated=True)
    )
    """
    DEPRECATED in 0.10, use build.verbose instead.
    """

    build_type: str = "Release"
    """
    The build type to use when building the project.

    Pre-defined CMake options are: ``Debug``, ``Release``, ``RelWithDebInfo``, ``MinSizeRel``

    Custom values can also be used.
    """

    source_dir: Path = Path()
    """
    The source directory to use when building the project.

    Currently only affects the native builder (not the setuptools plugin).
    """

    targets: Optional[List[str]] = dataclasses.field(
        default=None, metadata=SettingsFieldMetadata(deprecated=True)
    )
    """
    DEPRECATED in 0.10; use build.targets instead.
    """

    toolchain_file: Optional[Path] = dataclasses.field(
        default=None, metadata=SettingsFieldMetadata(override_only=True)
    )
    """
    The CMAKE_TOOLCHAIN_FILE / --toolchain used for cross-compilation.

    This is only allowed in overrides or config-settings.
    """

    python_hints: bool = True
    """
    Do not pass the current environment's python hints such as ``Python_EXECUTABLE``.
    Primarily used for cross-compilation where the CMAKE_TOOLCHAIN_FILE should handle it
    instead.
    """


@dataclasses.dataclass
class SearchSettings:
    site_packages: bool = True
    """
    Add the python build environment site_packages folder to the CMake prefix paths.
    """


@dataclasses.dataclass
class NinjaSettings:
    minimum_version: Optional[Version] = dataclasses.field(
        default=None, metadata=SettingsFieldMetadata(deprecated=True)
    )
    """
    DEPRECATED in 0.8; use version instead.
    """

    version: SpecifierSet = SpecifierSet(">=1.5")
    """
    The versions of Ninja to allow.

    If Ninja is not present on the system or does not pass this specifier, it will
    be downloaded via PyPI if possible with the equivalent specifier used.

    An empty string will disable this check.

    .. seealso::
       - :confval:`cmake.version`
       - :confval:`ninja.make-fallback`
    """

    make_fallback: bool = True
    """
    Use Make as a fallback if a suitable Ninja executable is not found.

    If Make is also not available on the system, a ninja dependency is added to the
    ``build-system.requires`` according to :confval:`ninja.version`.

    .. seealso::
       :confval:`ninja.version`
    """


@dataclasses.dataclass
class LoggingSettings:
    level: Literal["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = (
        "WARNING"
    )
    """
    The logging level to display.
    """


@dataclasses.dataclass
class SDistSettings:
    include: List[str] = dataclasses.field(default_factory=list)
    """
    Files to include in the SDist even if they are skipped by default. Supports gitignore syntax.

    Always takes precedence over :confval:`sdist.exclude`

    .. seealso::
       :confval:`sdist.exclude`
    """

    exclude: List[str] = dataclasses.field(default_factory=list)
    """
    Files to exclude from the SDist even if they are included by default. Supports gitignore syntax.

    .. seealso::
       :confval:`sdist.include`
    """

    inclusion_mode: Optional[Literal["classic", "default", "manual"]] = (
        dataclasses.field(
            default=None,
            metadata=SettingsFieldMetadata(display_default='"default"  # "classic"'),
        )
    )
    """
    Method to use to compute the files to include and exclude.

    The methods are:

    * "default": Process the git ignore files. Shortcuts on ignored directories.
    * "classic": The behavior before 0.12, like "default" but does not shortcut directories.
    * "manual": No extra logic, based on include/exclude only.

    If you don't set this, it will be "default" unless you set the minimum
    version below 0.12, in which case it will be "classic".

    .. versionadded: 0.12
    """

    reproducible: bool = True
    """
    Try to build a reproducible distribution.

    Unix and Python 3.9+ recommended.

    ``SOURCE_DATE_EPOCH`` will be used for timestamps, or a fixed value if not set.
    """

    cmake: bool = False
    """
    If set to True, CMake will be run before building the SDist.
    """


@dataclasses.dataclass
class WheelSettings:
    packages: Optional[Union[List[str], Dict[str, str]]] = dataclasses.field(
        default=None,
        metadata=SettingsFieldMetadata(
            display_default='["src/<package>", "python/<package>", "<package>"]'
        ),
    )
    """
    A list of packages to auto-copy into the wheel.

    If this is not set, it will default to the first of ``src/<package>``, ``python/<package>``, or
    ``<package>`` if they exist.  The prefix(s) will be stripped from the
    package name inside the wheel.

    If a dict, provides a mapping of package name to source directory.
    """

    py_api: str = ""
    """
    The Python version tag used in the wheel file.

    The default (empty string) will use the default Python version.

    You can also set this to "cp38" to enable the CPython 3.8+ Stable
    ABI / Limited API (only on CPython and if the version is sufficient,
    otherwise this has no effect). Or you can set it to "py3" or "py2.py3" to
    ignore Python ABI compatibility. The ABI tag is inferred from this tag.

    This value is used to construct ``SKBUILD_SABI_COMPONENT`` CMake variable.
    """

    expand_macos_universal_tags: bool = False
    """
    Fill out extra tags that are not required.

    This adds "x86_64" and "arm64" to the list of platforms when "universal2" is used,
    which helps older Pip's (before 21.0.1) find the correct wheel.
    """

    install_dir: str = ""
    """
    The CMake install prefix relative to the platlib wheel path.

    You might set this to the package name to install everything under the package namespace
    in a pythonic design.

    The original dir is still at ``SKBUILD_PLATLIB_DIR`` (also ``SKBUILD_DATA_DIR``, etc.
    are available).

    .. warning::
       EXPERIMENTAL An absolute path will be one level higher than the platlib
       root, giving access to "/platlib", "/data", "/headers", and "/scripts".
    """

    license_files: Optional[List[str]] = None
    """
    A list of license files to include in the wheel. Supports glob patterns.

    The default is ``["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"]``.

    .. warning::
       Must not be set if ``project.license-files`` is set.
    """

    cmake: bool = True
    """
    Run CMake as part of building the wheel.
    """

    platlib: Optional[bool] = None
    """
    Target the platlib or the purelib.

    If not set, the default is to target the platlib if :confval:`wheel.cmake` is ``true``,
    and the purelib otherwise.
    """

    exclude: List[str] = dataclasses.field(default_factory=list)
    """
    A set of patterns to exclude from the wheel.

    This is additive to the SDist exclude patterns. This applies to the final paths
    in the wheel, and can exclude files from CMake output as well.  Editable installs
    may not respect this exclusion.
    """

    build_tag: str = ""
    """
    The build tag to use for the wheel. If empty, no build tag is used.
    """

    tags: Optional[List[str]] = dataclasses.field(
        default=None,
        metadata=SettingsFieldMetadata(override_only=True),
    )
    """
    Wheel tags to manually force, {interpreter}-{abi}-{platform} format.

    Manually specify the wheel tags to use, ignoring other inputs such as
    ``wheel.py-api``. Each tag must be of the format
    {interpreter}-{abi}-{platform}.  If not specified, these tags are
    automatically calculated. This is only allowed in overrides or
    config-settings.
    """


@dataclasses.dataclass
class BackportSettings:
    find_python: Version = Version("3.26.1")
    """
    If CMake is less than this value, backport a copy of FindPython.

    Set to ``0`` or an empty string to disable this.
    """


@dataclasses.dataclass
class EditableSettings:
    mode: Literal["redirect", "inplace"] = "redirect"
    """
    Select the editable mode to use. Can be "redirect" (default) or "inplace".
    """

    verbose: bool = True
    """
    Turn on verbose output for the editable mode rebuilds.
    """

    rebuild: bool = False
    """
    Rebuild the project when the package is imported.

    :confval:`build-dir` must be set.
    """


@dataclasses.dataclass
class BuildSettings:
    tool_args: List[str] = dataclasses.field(default_factory=list)
    """
    Extra args to pass directly to the builder in the build step.
    """

    targets: List[str] = dataclasses.field(default_factory=list)
    """
    The build targets to use when building the project.

    If not specified or an empty list, the default target is used.
    """

    verbose: bool = False
    """
    Verbose printout when building.

    Equivalent to ``CMAKE_VERBOSE_MAKEFILE``.
    """

    requires: List[str] = dataclasses.field(default_factory=list)
    """
    Additional ``build-system.requires``.

    Intended to be used in combination with ``overrides``.
    """


@dataclasses.dataclass
class InstallSettings:
    components: List[str] = dataclasses.field(default_factory=list)
    """
    The components to install.

    If not specified or an empty list, all default components are installed.
    """

    strip: Optional[bool] = dataclasses.field(
        default=None, metadata=SettingsFieldMetadata(display_default="true")
    )
    """
    Whether to strip the binaries.

    Equivalent to ``--strip`` in ``cmake install``.

    True for release builds (`Release` or `MinSizeRel`) on scikit-build-core 0.5+.

    .. note::
       0.5-0.10.5 also incorrectly set this for debug builds.
    """


@dataclasses.dataclass
class GenerateSettings:
    path: Path
    """
    The path (relative to platlib) for the file to generate.
    """

    template: str = ""
    """
    The template string to use for the file.

    Template style placeholders are available for all the metadata.

    Either this or :confval:`generate[].template-path` must be set.

    .. seealso::
       :confval:`generate[].template-path`
    """

    template_path: Optional[Path] = None
    """
    The path to the template file. If empty, a template must be set.

    Either this or :confval:`generate[].template` must be set.

    .. seealso::
       :confval:`generate[].template`
    """

    location: Literal["install", "build", "source"] = "install"
    """
    The place to put the generated file.

    The ``build`` directory is useful for CMake files, and the ``install`` directory is
    useful for Python files, usually. You can also write directly to the ``source``
    directory, will overwrite existing files & remember to gitignore the file.
    """


@dataclasses.dataclass
class MessagesSettings:
    """
    Settings for messages.
    """

    after_failure: str = ""
    """
    A message to print after a build failure.
    """

    after_success: str = ""
    """
    A message to print after a successful build.
    """


@dataclasses.dataclass
class ScikitBuildSettings:
    cmake: CMakeSettings = dataclasses.field(default_factory=CMakeSettings)
    ninja: NinjaSettings = dataclasses.field(default_factory=NinjaSettings)
    logging: LoggingSettings = dataclasses.field(default_factory=LoggingSettings)
    sdist: SDistSettings = dataclasses.field(default_factory=SDistSettings)
    wheel: WheelSettings = dataclasses.field(default_factory=WheelSettings)
    backport: BackportSettings = dataclasses.field(default_factory=BackportSettings)
    editable: EditableSettings = dataclasses.field(default_factory=EditableSettings)
    build: BuildSettings = dataclasses.field(default_factory=BuildSettings)
    install: InstallSettings = dataclasses.field(default_factory=InstallSettings)
    generate: List[GenerateSettings] = dataclasses.field(default_factory=list)
    messages: MessagesSettings = dataclasses.field(default_factory=MessagesSettings)
    search: SearchSettings = dataclasses.field(default_factory=SearchSettings)

    metadata: Dict[str, Dict[str, Any]] = dataclasses.field(default_factory=dict)
    """
    List dynamic metadata fields and hook locations in this table.
    """

    strict_config: bool = True
    """
    Strictly check all config options.

    If False, warnings will be printed for unknown options.

    If True, an error will be raised.
    """

    experimental: bool = False
    """
    Enable early previews of features not finalized yet.
    """

    minimum_version: Optional[Version] = dataclasses.field(
        default=None,
        metadata=SettingsFieldMetadata(
            display_default='"{version}"  # current version'
        ),
    )
    """
    If set, this will provide a method for backward compatibility.
    """

    build_dir: str = ""
    """
    The CMake build directory. Defaults to a unique temporary directory.

    This can be set to reuse the build directory from previous runs.
    """

    fail: Optional[bool] = dataclasses.field(
        default=None,
        metadata=SettingsFieldMetadata(override_only=True),
    )
    """
    Immediately fail the build. This is only allowed in overrides or config-settings.
    """



from __future__ import annotations

import dataclasses
import enum
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from .._logging import rich_print

if TYPE_CHECKING:
    from collections.abc import Generator

__all__ = ["Token", "TokenType", "tokenize"]


def __dir__() -> list[str]:
    return __all__


TOKEN_EXPRS = {
    "BRACKET_COMMENT": r"\s*#\[(?P<bc1>=*)\[(?s:.)*?\](?P=bc1)\]",
    "COMMENT": r"#.*$",
    "QUOTED": r'"(?:\\(?s:.)|[^"\\])*?"',
    "BRACKET_QUOTE": r"\[(?P<bq1>=*)\[(?s:.)*?\](?P=bq1)\]",
    "OPEN_PAREN": r"\(",
    "CLOSE_PAREN": r"\)",
    "LEGACY": r'\b\w+=[^\s"()$\\]*(?:"[^"\\]*"[^\s"()$\\]*)*|"(?:[^"\\]*(?:\\.[^"\\]*)*)*"',
    "UNQUOTED": r"(?:\\.|[^\s()#\"\\])+",
}


class TokenType(enum.Enum):
    BRACKET_COMMENT = enum.auto()
    COMMENT = enum.auto()
    UNQUOTED = enum.auto()
    QUOTED = enum.auto()
    BRACKET_QUOTE = enum.auto()
    LEGACY = enum.auto()
    OPEN_PAREN = enum.auto()
    CLOSE_PAREN = enum.auto()
    WHITESPACE = enum.auto()


@dataclasses.dataclass(frozen=True)
class Token:
    __slots__ = ("start", "stop", "type", "value")

    type: TokenType
    start: int
    stop: int
    value: str

    def __str__(self) -> str:
        return f"{self.type.name}({self.value!r})"


def tokenize(contents: str) -> Generator[Token, None, None]:
    tok_regex = "|".join(f"(?P<{n}>{v})" for n, v in TOKEN_EXPRS.items())
    last = 0
    for match in re.finditer(tok_regex, contents, re.MULTILINE):
        for typ, value in match.groupdict().items():
            if typ in TOKEN_EXPRS and value is not None:
                if match.start() != last:
                    yield Token(
                        TokenType.WHITESPACE,
                        last,
                        match.start(),
                        contents[last : match.start()],
                    )
                last = match.end()
                yield Token(TokenType[typ], match.start(), match.end(), value)


if __name__ == "__main__":
    with Path(sys.argv[1]).open(encoding="utf-8-sig") as f:
        for token in tokenize(f.read()):
            rich_print(
                "{green}{token.type.name}{red}({default}{token.value}{red})",
                token=token,
            )



"""
This is the entry point for the build backend. Items in this module are designed for the build backend API.
"""

from __future__ import annotations

import sys

from .._compat import tomllib

__all__ = [
    "build_editable",
    "build_sdist",
    "build_wheel",
    "get_requires_for_build_editable",
    "get_requires_for_build_sdist",
    "get_requires_for_build_wheel",
    "prepare_metadata_for_build_editable",
    "prepare_metadata_for_build_wheel",
]


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
    metadata_directory: str | None = None,
) -> str:
    from .._logging import rich_print
    from ..errors import FailedLiveProcessError
    from .wheel import _build_wheel_impl

    try:
        return _build_wheel_impl(
            wheel_directory,
            config_settings,
            metadata_directory,
            editable=False,
        ).wheel_filename
    except FailedLiveProcessError as err:
        sys.stdout.flush()
        rich_print("\n{bold}***", *err.args, color="red", file=sys.stderr)
        if err.msg:
            rich_print(err.msg)
        raise SystemExit(1) from None


def build_editable(
    wheel_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
    metadata_directory: str | None = None,
) -> str:
    from .._logging import rich_print
    from ..errors import FailedLiveProcessError
    from .wheel import _build_wheel_impl

    try:
        return _build_wheel_impl(
            wheel_directory,
            config_settings,
            metadata_directory,
            editable=True,
        ).wheel_filename
    except FailedLiveProcessError as err:
        sys.stdout.flush()
        rich_print("\n{bold}***", *err.args, color="red", file=sys.stderr)
        if err.msg:
            rich_print(err.msg)
        raise SystemExit(1) from None


def _has_safe_metadata() -> bool:
    try:
        with open("pyproject.toml", "rb") as f:  # noqa: PTH123
            pyproject = tomllib.load(f)
    except FileNotFoundError:
        return True

    overrides = pyproject.get("tool", {}).get("scikit-build", {}).get("overrides", [])
    for override in overrides:
        if_override = override.get("if", {})
        if "failed" in if_override or "failed" in if_override.get("any", {}):
            return False

    return True


if _has_safe_metadata():

    def prepare_metadata_for_build_wheel(
        metadata_directory: str,
        config_settings: dict[str, list[str] | str] | None = None,
    ) -> str:
        """Prepare metadata for building a wheel. Does not build the wheel. Returns the dist-info directory."""
        from .wheel import _build_wheel_impl

        return _build_wheel_impl(
            None, config_settings, metadata_directory, editable=False
        ).wheel_filename  # actually returns the dist-info directory

    def prepare_metadata_for_build_editable(
        metadata_directory: str,
        config_settings: dict[str, list[str] | str] | None = None,
    ) -> str:
        """Prepare metadata for building a wheel. Does not build the wheel. Returns the dist-info directory."""

        from .wheel import _build_wheel_impl

        return _build_wheel_impl(
            None, config_settings, metadata_directory, editable=True
        ).wheel_filename  # actually returns the dist-info directory

    __all__ += [
        "prepare_metadata_for_build_editable",
        "prepare_metadata_for_build_wheel",
    ]


def build_sdist(
    sdist_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
) -> str:
    from .._logging import rich_print
    from ..errors import FailedLiveProcessError
    from .sdist import build_sdist as skbuild_build_sdist

    try:
        return skbuild_build_sdist(sdist_directory, config_settings)
    except FailedLiveProcessError as err:
        sys.stdout.flush()
        rich_print("\n{bold}***", *err.args, color="red", file=sys.stderr)
        if err.msg:
            rich_print(err.msg)
        raise SystemExit(1) from None


def get_requires_for_build_sdist(
    config_settings: dict[str, str | list[str]] | None = None,
) -> list[str]:
    from ..builder.get_requires import GetRequires

    requires = GetRequires.from_config_settings(config_settings)

    # These are only injected if cmake is required for the SDist step
    cmake_requires = (
        [*requires.cmake(), *requires.ninja()] if requires.settings.sdist.cmake else []
    )

    return [
        *cmake_requires,
        *requires.dynamic_metadata(),
    ]


def get_requires_for_build_wheel(
    config_settings: dict[str, str | list[str]] | None = None,
) -> list[str]:
    from ..builder.get_requires import GetRequires

    requires = GetRequires.from_config_settings(config_settings)

    # These are only injected if cmake is required for the wheel step
    cmake_requires = (
        [*requires.cmake(), *requires.ninja()] if requires.settings.wheel.cmake else []
    )

    return [
        *cmake_requires,
        *requires.dynamic_metadata(),
    ]


def get_requires_for_build_editable(
    config_settings: dict[str, str | list[str]] | None = None,
) -> list[str]:
    from ..builder.get_requires import GetRequires

    requires = GetRequires.from_config_settings(config_settings)

    # These are only injected if cmake is required for the wheel step
    cmake_requires = (
        [*requires.cmake(), *requires.ninja()] if requires.settings.wheel.cmake else []
    )

    return [
        *cmake_requires,
        *requires.dynamic_metadata(),
    ]



import builtins
import dataclasses
import json
import sys
import typing
from pathlib import Path
from typing import Any, Callable, Dict, List, Type, TypeVar, Union  # noqa: TID251

from .._compat.builtins import ExceptionGroup
from .._compat.typing import get_args, get_origin
from .model.cache import Cache
from .model.cmakefiles import CMakeFiles
from .model.codemodel import CodeModel, Target
from .model.directory import Directory
from .model.index import Index

__all__ = ["load_reply_dir"]


def __dir__() -> List[str]:
    return __all__


T = TypeVar("T")

InputDict = Dict[str, Any]


class Converter:
    def __init__(self, base_dir: Path) -> None:
        self.base_dir = base_dir

    def load(self) -> Index:
        """
        Load the newest index.json file and return the Index object.
        """
        index_file = sorted(self.base_dir.glob("index-*"))[-1]
        with index_file.open(encoding="utf-8") as f:
            data = json.load(f)

        return self.make_class(data, Index)

    def _load_from_json(self, name: Path, target: Type[T]) -> T:
        with self.base_dir.joinpath(name).open(encoding="utf-8") as f:
            data = json.load(f)

        return self.make_class(data, target)

    def make_class(self, data: InputDict, target: Type[T]) -> T:
        """
        Convert a dict to a dataclass. Automatically load a few nested jsonFile classes.
        """
        if (
            target in {CodeModel, Target, Cache, CMakeFiles, Directory}
            and "jsonFile" in data
            and data["jsonFile"] is not None
        ):
            return self._load_from_json(Path(data["jsonFile"]), target)

        input_dict: Dict[str, Type[Any]] = {}
        exceptions: List[Exception] = []

        # We don't have DataclassInstance exposed in typing yet
        for field in dataclasses.fields(target):  # type: ignore[arg-type]
            json_field = field.name.replace("_v", "-v").replace(
                "cmakefiles", "cmakeFiles"
            )
            if json_field in data:
                field_type = field.type
                try:
                    input_dict[field.name] = self._convert_any(
                        data[json_field], field_type
                    )
                except TypeError as err:
                    msg = f"Failed to convert field {field.name!r} of type {field_type}"
                    if sys.version_info < (3, 11):
                        err.__notes__ = [*getattr(err, "__notes__", []), msg]  # type: ignore[attr-defined]
                    else:
                        err.add_note(msg)  # pylint: disable=no-member
                    exceptions.append(err)
                except ExceptionGroup as err:
                    exceptions.append(err)

        if exceptions:
            msg = f"Failed converting {target}"
            raise ExceptionGroup(msg, exceptions)

        return target(**input_dict)

    @typing.overload
    def _convert_any(self, item: Any, target: Type[T]) -> T: ...
    @typing.overload
    def _convert_any(self, item: Any, target: Any) -> Any: ...

    def _convert_any(self, item: Any, target: Union[Type[T], Any]) -> Any:
        if dataclasses.is_dataclass(target) and isinstance(target, type):
            # We don't have DataclassInstance exposed in typing yet
            return self.make_class(item, target)
        origin = get_origin(target)
        if origin is not None:
            if origin is list:
                return [self._convert_any(i, get_args(target)[0]) for i in item]
            if origin is Union:
                return self._convert_any(item, get_args(target)[0])

        return target(item)  # type: ignore[call-arg]


def load_reply_dir(path: Path) -> Index:
    return Converter(path).load()


if __name__ == "__main__":
    import argparse

    rich_print: Callable[[object], None]
    try:
        from rich import print as rich_print
    except ModuleNotFoundError:
        rich_print = builtins.print

    parser = argparse.ArgumentParser(
        prog="python -m scikit_build_core.file_api.reply",
        allow_abbrev=False,
        description="Read a query written out to a build directory.",
    )
    parser.add_argument("reply_dir", type=Path, help="Path to the reply directory")
    args = parser.parse_args()

    reply = Path(args.reply_dir)
    rich_print(load_reply_dir(reply))



import dataclasses
from pathlib import Path
from typing import List, Optional, Union

from .common import Paths

__all__ = ["BacktraceGraph", "Directory", "InstallRule", "Node", "Target"]


def __dir__() -> List[str]:
    return __all__


@dataclasses.dataclass(frozen=True)
class Target:
    id: str
    index: int


@dataclasses.dataclass(frozen=True)
class InstallRule:
    component: str
    type = str
    destination: Optional[Path] = None
    paths: List[Union[str, Paths]] = dataclasses.field(default_factory=list)
    isExcludeFromAll: bool = False
    isForAllComponents: bool = False
    isOptional: bool = False
    targetId: Optional[str] = None
    targetIndex: Optional[int] = None
    targetIsImportLibrary: bool = False
    targetInstallNameLink: Optional[str] = None
    exportName: Optional[str] = None
    exportTargets: List[Target] = dataclasses.field(default_factory=list)
    runtimeDependencySetName: Optional[str] = None
    runtimeDependencySetType: Optional[str] = None
    fileSetName: Optional[str] = None
    fileSetType: Optional[str] = None
    fileSetDirectories: List[Path] = dataclasses.field(default_factory=list)
    fileSetTarget: Optional[Target] = None
    scriptFile: Optional[Path] = None
    backtrace: Optional[int] = None


@dataclasses.dataclass(frozen=True)
class Node:
    file: int
    line: Optional[int] = None
    command: Optional[int] = None
    parent: Optional[int] = None


@dataclasses.dataclass(frozen=True)
class BacktraceGraph:
    nodes: List[Node]
    commands: List[str]
    files: List[Path]


@dataclasses.dataclass(frozen=True)
class Directory:
    paths: Paths
    installers: List[InstallRule]
    backtraceGraph: BacktraceGraph



"""
Make documentation for the skbuild model in sphinx format.
"""

from __future__ import annotations

import dataclasses
import textwrap
import typing
from collections import OrderedDict

from .documentation import mk_docs
from .skbuild_model import ScikitBuildSettings

if typing.TYPE_CHECKING:
    from .documentation import DCDoc

__all__ = ["mk_skbuild_docs"]


def __dir__() -> list[str]:
    return __all__


@dataclasses.dataclass
class Document:
    sections: dict[str, Section] = dataclasses.field(
        default_factory=OrderedDict, init=False
    )

    def format(self) -> str:
        return "\n".join(self.sections[it].format() for it in sorted(self.sections))


@dataclasses.dataclass
class Section:
    TEMPLATE: typing.ClassVar[str] = textwrap.dedent("""\
    {section_level} {section}

    {content}""")
    name: str
    level: int = 2
    items: dict[str, Item] = dataclasses.field(default_factory=OrderedDict, init=False)

    def format(self) -> str:
        return self.TEMPLATE.format(
            section=self.name,
            section_level="#" * self.level,
            content="\n".join(self.items[it].format() for it in sorted(self.items)),
        )


@dataclasses.dataclass
class Item:
    TEMPLATE: typing.ClassVar[str] = textwrap.dedent("""\
    ```{{eval-rst}}
    .. confval:: {item.name}
      :type: ``{item.type}``{sphinx_default}

      {docs}
    ```
    """)
    item: DCDoc

    def ignore_default(self) -> bool:
        """
        Check if we want to ignore displaying the default
        """
        return self.item.default in ('""', "[]", "{}")

    def default(self) -> str:
        """
        Formatted text that includes the `:default:` key or not
        """
        if self.ignore_default():
            return ""
        return f"\n  :default: {self.item.default}"

    def format(self) -> str:
        # Replace all new-lines with appropriately rst indented lines
        docs = self.item.docs.replace("\n", "\n  ")
        # Replace all lines with pure whitespace to satisfy pre-commit
        docs = docs.replace("\n  \n", "\n\n")
        return self.TEMPLATE.format(
            item=self.item,
            sphinx_default=self.default(),
            docs=docs,
        )


def mk_skbuild_docs() -> str:
    doc = Document()
    for doc_item in mk_docs(ScikitBuildSettings):
        name_split = doc_item.name.rsplit(".", maxsplit=1)
        section_name = f"{name_split[0]}" if len(name_split) > 1 else "(top-level)"
        section = doc.sections.setdefault(section_name, Section(section_name))
        section.items[doc_item.name] = Item(doc_item)
    return doc.format()


if __name__ == "__main__":
    print(mk_skbuild_docs())  # noqa: T201



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

import setuptools

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

from .._compat.typing import TypeVar

__all__ = ["setup"]


def __dir__() -> list[str]:
    return __all__


_DistributionT = TypeVar(
    "_DistributionT",
    bound="setuptools._distutils.dist.Distribution",
    default=setuptools.Distribution,
)


def setup(
    *,
    cmake_args: Sequence[str] = (),
    cmake_install_dir: str = "",
    cmake_source_dir: str = "",
    cmake_with_sdist: bool = False,
    cmake_languages: Sequence[str] | None = None,
    cmake_minimum_required_version: str | None = None,
    cmake_process_manifest_hook: Callable[[list[str]], list[str]] | None = None,
    cmake_install_target: str = "install",
    distclass: type[_DistributionT] = setuptools.Distribution,  # type: ignore[assignment]
    **kw: Any,
) -> _DistributionT:
    assert not cmake_install_dir, "cmake_install_dir not supported yet"
    assert not cmake_with_sdist, "cmake_with_sdist not supported yet"
    assert cmake_process_manifest_hook is None, (
        "cmake_process_manifest_hook not supported yet"
    )
    assert cmake_install_target == "install", "cmake_install_target not supported yet"

    if cmake_languages is not None:
        warnings.warn("cmake_languages no longer has any effect", stacklevel=2)

    if cmake_minimum_required_version is not None:
        warnings.warn("Set via pyproject.toml", stacklevel=2)

    return setuptools.setup(
        cmake_source_dir=cmake_source_dir,
        cmake_args=cmake_args,
        distclass=distclass,
        **kw,
    )



from __future__ import annotations

import sys
import typing

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_args, get_origin
else:
    from typing import Annotated, get_args, get_origin

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

if sys.version_info < (3, 11):
    if typing.TYPE_CHECKING:
        from typing_extensions import Self, assert_never
    else:
        Self = object

        def assert_never(_: object) -> None:
            msg = "Expected code to be unreachable"
            raise AssertionError(msg)
else:
    from typing import Self, assert_never

if sys.version_info < (3, 13):
    if typing.TYPE_CHECKING:
        from typing_extensions import TypeVar
    else:
        # The final noqa is a false positive, see https://github.com/astral-sh/ruff/issues/22178
        def TypeVar(  # noqa: N802
            *args: object,
            default: object = None,  # noqa: ARG001
            **kwargs: object,
        ) -> typing.TypeVar:
            return typing.TypeVar(*args, **kwargs)
else:
    from typing import TypeVar


__all__ = [
    "Annotated",
    "Self",
    "TypeAlias",
    "TypeVar",
    "assert_never",
    "get_args",
    "get_origin",
]


def __dir__() -> list[str]:
    return __all__
